# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import grpc

from chalk._gen.chalk.server.v1 import monitoring_pb2 as chalk_dot_server_dot_v1_dot_monitoring__pb2


class MonitoringServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.TestPagerDutyIntegration = channel.unary_unary(
            "/chalk.server.v1.MonitoringService/TestPagerDutyIntegration",
            request_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.TestPagerDutyIntegrationRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.TestPagerDutyIntegrationResponse.FromString,
        )
        self.AddPagerDutyIntegration = channel.unary_unary(
            "/chalk.server.v1.MonitoringService/AddPagerDutyIntegration",
            request_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.AddPagerDutyIntegrationRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.AddPagerDutyIntegrationResponse.FromString,
        )
        self.DeletePagerDutyIntegration = channel.unary_unary(
            "/chalk.server.v1.MonitoringService/DeletePagerDutyIntegration",
            request_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.DeletePagerDutyIntegrationRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.DeletePagerDutyIntegrationResponse.FromString,
        )
        self.SetDefaultPagerDutyIntegration = channel.unary_unary(
            "/chalk.server.v1.MonitoringService/SetDefaultPagerDutyIntegration",
            request_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.SetDefaultPagerDutyIntegrationRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.SetDefaultPagerDutyIntegrationResponse.FromString,
        )
        self.UpdatePagerDutyIntegration = channel.unary_unary(
            "/chalk.server.v1.MonitoringService/UpdatePagerDutyIntegration",
            request_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.UpdatePagerDutyIntegrationRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.UpdatePagerDutyIntegrationResponse.FromString,
        )
        self.GetAllPagerDutyIntegrations = channel.unary_unary(
            "/chalk.server.v1.MonitoringService/GetAllPagerDutyIntegrations",
            request_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.GetAllPagerDutyIntegrationsRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.GetAllPagerDutyIntegrationsResponse.FromString,
        )
        self.GetPagerDutyIntegration = channel.unary_unary(
            "/chalk.server.v1.MonitoringService/GetPagerDutyIntegration",
            request_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.GetPagerDutyIntegrationRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.GetPagerDutyIntegrationResponse.FromString,
        )
        self.TestIncidentIoIntegration = channel.unary_unary(
            "/chalk.server.v1.MonitoringService/TestIncidentIoIntegration",
            request_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.TestIncidentIoIntegrationRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.TestIncidentIoIntegrationResponse.FromString,
        )
        self.AddIncidentIoIntegration = channel.unary_unary(
            "/chalk.server.v1.MonitoringService/AddIncidentIoIntegration",
            request_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.AddIncidentIoIntegrationRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.AddIncidentIoIntegrationResponse.FromString,
        )
        self.DeleteIncidentIoIntegration = channel.unary_unary(
            "/chalk.server.v1.MonitoringService/DeleteIncidentIoIntegration",
            request_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.DeleteIncidentIoIntegrationRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.DeleteIncidentIoIntegrationResponse.FromString,
        )
        self.UpdateIncidentIoIntegration = channel.unary_unary(
            "/chalk.server.v1.MonitoringService/UpdateIncidentIoIntegration",
            request_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.UpdateIncidentIoIntegrationRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.UpdateIncidentIoIntegrationResponse.FromString,
        )
        self.GetAllIncidentIoIntegrations = channel.unary_unary(
            "/chalk.server.v1.MonitoringService/GetAllIncidentIoIntegrations",
            request_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.GetAllIncidentIoIntegrationsRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.GetAllIncidentIoIntegrationsResponse.FromString,
        )
        self.GetIncidentIoIntegration = channel.unary_unary(
            "/chalk.server.v1.MonitoringService/GetIncidentIoIntegration",
            request_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.GetIncidentIoIntegrationRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.GetIncidentIoIntegrationResponse.FromString,
        )


class MonitoringServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def TestPagerDutyIntegration(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def AddPagerDutyIntegration(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def DeletePagerDutyIntegration(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def SetDefaultPagerDutyIntegration(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def UpdatePagerDutyIntegration(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetAllPagerDutyIntegrations(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetPagerDutyIntegration(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def TestIncidentIoIntegration(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def AddIncidentIoIntegration(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def DeleteIncidentIoIntegration(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def UpdateIncidentIoIntegration(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetAllIncidentIoIntegrations(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetIncidentIoIntegration(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_MonitoringServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "TestPagerDutyIntegration": grpc.unary_unary_rpc_method_handler(
            servicer.TestPagerDutyIntegration,
            request_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.TestPagerDutyIntegrationRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.TestPagerDutyIntegrationResponse.SerializeToString,
        ),
        "AddPagerDutyIntegration": grpc.unary_unary_rpc_method_handler(
            servicer.AddPagerDutyIntegration,
            request_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.AddPagerDutyIntegrationRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.AddPagerDutyIntegrationResponse.SerializeToString,
        ),
        "DeletePagerDutyIntegration": grpc.unary_unary_rpc_method_handler(
            servicer.DeletePagerDutyIntegration,
            request_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.DeletePagerDutyIntegrationRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.DeletePagerDutyIntegrationResponse.SerializeToString,
        ),
        "SetDefaultPagerDutyIntegration": grpc.unary_unary_rpc_method_handler(
            servicer.SetDefaultPagerDutyIntegration,
            request_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.SetDefaultPagerDutyIntegrationRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.SetDefaultPagerDutyIntegrationResponse.SerializeToString,
        ),
        "UpdatePagerDutyIntegration": grpc.unary_unary_rpc_method_handler(
            servicer.UpdatePagerDutyIntegration,
            request_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.UpdatePagerDutyIntegrationRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.UpdatePagerDutyIntegrationResponse.SerializeToString,
        ),
        "GetAllPagerDutyIntegrations": grpc.unary_unary_rpc_method_handler(
            servicer.GetAllPagerDutyIntegrations,
            request_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.GetAllPagerDutyIntegrationsRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.GetAllPagerDutyIntegrationsResponse.SerializeToString,
        ),
        "GetPagerDutyIntegration": grpc.unary_unary_rpc_method_handler(
            servicer.GetPagerDutyIntegration,
            request_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.GetPagerDutyIntegrationRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.GetPagerDutyIntegrationResponse.SerializeToString,
        ),
        "TestIncidentIoIntegration": grpc.unary_unary_rpc_method_handler(
            servicer.TestIncidentIoIntegration,
            request_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.TestIncidentIoIntegrationRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.TestIncidentIoIntegrationResponse.SerializeToString,
        ),
        "AddIncidentIoIntegration": grpc.unary_unary_rpc_method_handler(
            servicer.AddIncidentIoIntegration,
            request_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.AddIncidentIoIntegrationRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.AddIncidentIoIntegrationResponse.SerializeToString,
        ),
        "DeleteIncidentIoIntegration": grpc.unary_unary_rpc_method_handler(
            servicer.DeleteIncidentIoIntegration,
            request_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.DeleteIncidentIoIntegrationRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.DeleteIncidentIoIntegrationResponse.SerializeToString,
        ),
        "UpdateIncidentIoIntegration": grpc.unary_unary_rpc_method_handler(
            servicer.UpdateIncidentIoIntegration,
            request_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.UpdateIncidentIoIntegrationRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.UpdateIncidentIoIntegrationResponse.SerializeToString,
        ),
        "GetAllIncidentIoIntegrations": grpc.unary_unary_rpc_method_handler(
            servicer.GetAllIncidentIoIntegrations,
            request_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.GetAllIncidentIoIntegrationsRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.GetAllIncidentIoIntegrationsResponse.SerializeToString,
        ),
        "GetIncidentIoIntegration": grpc.unary_unary_rpc_method_handler(
            servicer.GetIncidentIoIntegration,
            request_deserializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.GetIncidentIoIntegrationRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_monitoring__pb2.GetIncidentIoIntegrationResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler("chalk.server.v1.MonitoringService", rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class MonitoringService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def TestPagerDutyIntegration(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.MonitoringService/TestPagerDutyIntegration",
            chalk_dot_server_dot_v1_dot_monitoring__pb2.TestPagerDutyIntegrationRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_monitoring__pb2.TestPagerDutyIntegrationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def AddPagerDutyIntegration(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.MonitoringService/AddPagerDutyIntegration",
            chalk_dot_server_dot_v1_dot_monitoring__pb2.AddPagerDutyIntegrationRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_monitoring__pb2.AddPagerDutyIntegrationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def DeletePagerDutyIntegration(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.MonitoringService/DeletePagerDutyIntegration",
            chalk_dot_server_dot_v1_dot_monitoring__pb2.DeletePagerDutyIntegrationRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_monitoring__pb2.DeletePagerDutyIntegrationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def SetDefaultPagerDutyIntegration(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.MonitoringService/SetDefaultPagerDutyIntegration",
            chalk_dot_server_dot_v1_dot_monitoring__pb2.SetDefaultPagerDutyIntegrationRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_monitoring__pb2.SetDefaultPagerDutyIntegrationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def UpdatePagerDutyIntegration(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.MonitoringService/UpdatePagerDutyIntegration",
            chalk_dot_server_dot_v1_dot_monitoring__pb2.UpdatePagerDutyIntegrationRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_monitoring__pb2.UpdatePagerDutyIntegrationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetAllPagerDutyIntegrations(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.MonitoringService/GetAllPagerDutyIntegrations",
            chalk_dot_server_dot_v1_dot_monitoring__pb2.GetAllPagerDutyIntegrationsRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_monitoring__pb2.GetAllPagerDutyIntegrationsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetPagerDutyIntegration(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.MonitoringService/GetPagerDutyIntegration",
            chalk_dot_server_dot_v1_dot_monitoring__pb2.GetPagerDutyIntegrationRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_monitoring__pb2.GetPagerDutyIntegrationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def TestIncidentIoIntegration(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.MonitoringService/TestIncidentIoIntegration",
            chalk_dot_server_dot_v1_dot_monitoring__pb2.TestIncidentIoIntegrationRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_monitoring__pb2.TestIncidentIoIntegrationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def AddIncidentIoIntegration(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.MonitoringService/AddIncidentIoIntegration",
            chalk_dot_server_dot_v1_dot_monitoring__pb2.AddIncidentIoIntegrationRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_monitoring__pb2.AddIncidentIoIntegrationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def DeleteIncidentIoIntegration(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.MonitoringService/DeleteIncidentIoIntegration",
            chalk_dot_server_dot_v1_dot_monitoring__pb2.DeleteIncidentIoIntegrationRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_monitoring__pb2.DeleteIncidentIoIntegrationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def UpdateIncidentIoIntegration(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.MonitoringService/UpdateIncidentIoIntegration",
            chalk_dot_server_dot_v1_dot_monitoring__pb2.UpdateIncidentIoIntegrationRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_monitoring__pb2.UpdateIncidentIoIntegrationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetAllIncidentIoIntegrations(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.MonitoringService/GetAllIncidentIoIntegrations",
            chalk_dot_server_dot_v1_dot_monitoring__pb2.GetAllIncidentIoIntegrationsRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_monitoring__pb2.GetAllIncidentIoIntegrationsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetIncidentIoIntegration(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.MonitoringService/GetIncidentIoIntegration",
            chalk_dot_server_dot_v1_dot_monitoring__pb2.GetIncidentIoIntegrationRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_monitoring__pb2.GetIncidentIoIntegrationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



import datetime as dt
import json
from typing import Any, Dict, List, Optional, Union

from google.protobuf import message as proto_message

from chalk._gen.chalk.common.v1 import chalk_error_pb2, online_query_pb2
from chalk.client import ChalkError, ChalkException, ErrorCode, ErrorCodeCategory, OnlineQueryContext, QueryMeta
from chalk.client._internal_models.models import INDEX_COL_NAME, OBSERVED_AT_COL_NAME, PKEY_COL_NAME, TS_COL_NAME
from chalk.client.models import (
    BulkOnlineQueryResponse,
    BulkOnlineQueryResult,
    BulkUploadFeaturesResult,
    FeatureResolutionMeta,
    FeatureResult,
    OnlineQueryManyRequest,
    OnlineQueryRequest,
    OnlineQueryResponse,
    OnlineQueryResultFeather,
)
from chalk.client.serialization.constants import GRPC_RESULT_METADATA_COL_PREFIX
from chalk.features._encoding.converter import PrimitiveFeatureConverter
from chalk.features._encoding.json import FeatureEncodingOptions
from chalk.parsed._proto.utils import (
    datetime_to_proto_timestamp,
    proto_duration_to_timedelta,
    proto_value_to_python,
    seconds_to_proto_duration,
    value_to_proto,
)
from chalk.utils.collections import BidirectionalMap
from chalk.utils.df_utils import arrow_ipc_to_record_batch, pa_table_to_pl_df, record_batch_to_arrow_ipc


def check_has_field(proto: proto_message.Message, field: str):
    if not proto.HasField(field):
        raise ValueError(f"Proto is missing expected field '{field}'.")


def get_field_or(proto: proto_message.Message, field_name: str, default_value: Any = None):
    if not proto.HasField(field_name):
        return default_value
    return getattr(proto, field_name)


class ChalkErrorConverter:
    _error_code_map: "BidirectionalMap[ErrorCode, chalk_error_pb2.ErrorCode]" = BidirectionalMap(
        [
            (ErrorCode.INTERNAL_SERVER_ERROR, chalk_error_pb2.ERROR_CODE_INTERNAL_SERVER_ERROR_UNSPECIFIED),
            (ErrorCode.PARSE_FAILED, chalk_error_pb2.ERROR_CODE_PARSE_FAILED),
            (ErrorCode.RESOLVER_NOT_FOUND, chalk_error_pb2.ERROR_CODE_RESOLVER_NOT_FOUND),
            (ErrorCode.INVALID_QUERY, chalk_error_pb2.ERROR_CODE_INVALID_QUERY),
            (ErrorCode.VALIDATION_FAILED, chalk_error_pb2.ERROR_CODE_VALIDATION_FAILED),
            (ErrorCode.RESOLVER_FAILED, chalk_error_pb2.ERROR_CODE_RESOLVER_FAILED),
            (ErrorCode.RESOLVER_TIMED_OUT, chalk_error_pb2.ERROR_CODE_RESOLVER_TIMED_OUT),
            (ErrorCode.UPSTREAM_FAILED, chalk_error_pb2.ERROR_CODE_UPSTREAM_FAILED),
            (ErrorCode.UNAUTHENTICATED, chalk_error_pb2.ERROR_CODE_UNAUTHENTICATED),
            (ErrorCode.UNAUTHORIZED, chalk_error_pb2.ERROR_CODE_UNAUTHORIZED),
            (ErrorCode.CANCELLED, chalk_error_pb2.ERROR_CODE_CANCELLED),
            (ErrorCode.DEADLINE_EXCEEDED, chalk_error_pb2.ERROR_CODE_DEADLINE_EXCEEDED),
        ]
    )

    _error_category_map: "BidirectionalMap[ErrorCodeCategory, chalk_error_pb2.ErrorCodeCategory]" = BidirectionalMap(
        [
            (ErrorCodeCategory.NETWORK, chalk_error_pb2.ERROR_CODE_CATEGORY_NETWORK_UNSPECIFIED),
            (ErrorCodeCategory.FIELD, chalk_error_pb2.ERROR_CODE_CATEGORY_FIELD),
            (ErrorCodeCategory.REQUEST, chalk_error_pb2.ERROR_CODE_CATEGORY_REQUEST),
        ]
    )

    @staticmethod
    def _chalk_error_code_encode(
        error_code: ErrorCode,
    ) -> chalk_error_pb2.ErrorCode:
        code_proto: Optional[chalk_error_pb2.ErrorCode] = ChalkErrorConverter._error_code_map.get(error_code)
        if code_proto is None:
            raise ValueError(f"Unrecognized error code: {error_code}")
        else:
            return code_proto

    @staticmethod
    def _chalk_error_code_decode(
        error_code_proto: chalk_error_pb2.ErrorCode,
    ) -> ErrorCode:
        code_py: Optional[ErrorCode] = ChalkErrorConverter._error_code_map.get_reverse(error_code_proto)
        if code_py is None:
            raise ValueError(f"Unrecognized error code protobuf: {error_code_proto}")
        else:
            return code_py

    @staticmethod
    def _chalk_error_code_category_encode(
        category_py: ErrorCodeCategory,
    ) -> chalk_error_pb2.ErrorCodeCategory:
        category_proto: Optional[chalk_error_pb2.ErrorCodeCategory] = ChalkErrorConverter._error_category_map.get(
            category_py
        )
        if category_proto is None:
            raise ValueError(f"Unrecognized error code: {category_proto}")
        else:
            return category_proto

    @staticmethod
    def _chalk_error_code_category_decode(
        category_proto: chalk_error_pb2.ErrorCodeCategory,
    ) -> ErrorCodeCategory:
        category_py: Optional[ErrorCodeCategory] = ChalkErrorConverter._error_category_map.get_reverse(category_proto)
        if category_py is None:
            raise ValueError(f"Unrecognized error code: {category_proto}")
        else:
            return category_py

    @staticmethod
    def chalk_exception_encode(
        exception: ChalkException,
    ) -> chalk_error_pb2.ChalkException:
        proto_exception = chalk_error_pb2.ChalkException()
        proto_exception.kind = exception.kind
        proto_exception.message = exception.message
        proto_exception.stacktrace = exception.stacktrace
        if exception.internal_stacktrace is not None:
            proto_exception.internal_stacktrace = exception.internal_stacktrace
        return proto_exception

    @staticmethod
    def chalk_exception_decode(exception_proto: chalk_error_pb2.ChalkException) -> ChalkException:
        return ChalkException.create(
            kind=exception_proto.kind,
            message=exception_proto.message,
            stacktrace=exception_proto.stacktrace,
            internal_stacktrace=exception_proto.internal_stacktrace or None,
        )

    @staticmethod
    def chalk_error_encode(
        error: ChalkError,
    ) -> chalk_error_pb2.ChalkError:
        proto_error = chalk_error_pb2.ChalkError()
        proto_error.code = ChalkErrorConverter._chalk_error_code_encode(error.code)
        proto_error.category = ChalkErrorConverter._chalk_error_code_category_encode(error.category)
        proto_error.message = error.message
        if error.display_primary_key:
            proto_error.display_primary_key = error.display_primary_key
        if error.display_primary_key_fqn:
            proto_error.display_primary_key_fqn = error.display_primary_key_fqn
        if error.exception:
            proto_error.exception.CopyFrom(ChalkErrorConverter.chalk_exception_encode(error.exception))
        if error.feature:
            proto_error.feature = error.feature
        if error.resolver:
            proto_error.resolver = error.resolver
        return proto_error

    @staticmethod
    def chalk_error_decode(
        error_proto: chalk_error_pb2.ChalkError,
    ) -> ChalkError:
        if error_proto.HasField("exception"):
            chalk_exception = ChalkErrorConverter.chalk_exception_decode(error_proto.exception)
        else:
            chalk_exception = None
        return ChalkError.create(
            code=ChalkErrorConverter._chalk_error_code_decode(error_proto.code),
            category=ChalkErrorConverter._chalk_error_code_category_decode(error_proto.category),
            message=error_proto.message,
            display_primary_key=get_field_or(error_proto, "display_primary_key", None),
            display_primary_key_fqn=get_field_or(error_proto, "display_primary_key_fqn", None),
            exception=chalk_exception,
            feature=get_field_or(error_proto, "feature", None),
            resolver=get_field_or(error_proto, "resolver", None),
        )


class UploadFeaturesBulkConverter:
    @staticmethod
    def upload_features_bulk_response_decode(
        response_proto: online_query_pb2.UploadFeaturesBulkResponse, trace_id: Optional[str]
    ) -> BulkUploadFeaturesResult:
        return BulkUploadFeaturesResult(
            errors=[ChalkErrorConverter.chalk_error_decode(e) for e in response_proto.errors],
            trace_id=trace_id,
        )


class OnlineQueryConverter:
    @staticmethod
    def _options_bool_field(online_query_context: online_query_pb2.OnlineQueryContext, field: str) -> Optional[bool]:
        unacceptable_has_fields = ("number_value", "string_value", "struct_value", "list_value")
        value = online_query_context.options[field]
        if value.HasField("null_value"):
            return None
        for fieldx in unacceptable_has_fields:
            if value.HasField(fieldx):
                raise ValueError(f"Expected bool_value in online_query_context.options.{field}, got {fieldx}")
        return value.bool_value

    @staticmethod
    def _online_query_context_decode(
        online_query_context: online_query_pb2.OnlineQueryContext,
    ) -> OnlineQueryContext:
        # Note that the proto online query context tends to differ from the python model. This function needs
        # frequent updates as more stuff gets moved into the python OnlineQueryContext
        return OnlineQueryContext(
            environment=online_query_context.environment,
            tags=[tag for tag in online_query_context.tags],
            required_resolver_tags=[tag for tag in online_query_context.required_resolver_tags],
        )

    @staticmethod
    def _feature_encoding_options_decode(
        feature_encoding_options: online_query_pb2.FeatureEncodingOptions,
    ) -> FeatureEncodingOptions:
        return FeatureEncodingOptions(
            encode_structs_as_objects=feature_encoding_options.encode_structs_as_objects,
        )

    @staticmethod
    def _query_meta_encode(query_meta: QueryMeta) -> online_query_pb2.OnlineQueryMetadata:
        proto_metadata = online_query_pb2.OnlineQueryMetadata()
        proto_metadata.execution_duration.CopyFrom(seconds_to_proto_duration(query_meta.execution_duration_s))
        if query_meta.deployment_id is not None:
            proto_metadata.deployment_id = query_meta.deployment_id
        if query_meta.environment_id is not None:
            proto_metadata.environment_id = query_meta.environment_id
        if query_meta.environment_name is not None:
            proto_metadata.environment_name = query_meta.environment_name
        if query_meta.query_id is not None:
            proto_metadata.query_id = query_meta.query_id
        if query_meta.query_timestamp is not None:
            proto_metadata.query_timestamp.CopyFrom(datetime_to_proto_timestamp(query_meta.query_timestamp))
        if query_meta.query_hash is not None:
            proto_metadata.query_hash = query_meta.query_hash
        if query_meta.explain_output is not None:
            proto_metadata.explain_output.CopyFrom(
                online_query_pb2.QueryExplainInfo(plan_string=query_meta.explain_output)
            )
        return proto_metadata

    @staticmethod
    def _query_meta_decode(meta_proto: online_query_pb2.OnlineQueryMetadata) -> QueryMeta:
        check_has_field(meta_proto, "execution_duration")
        execution_duration = proto_duration_to_timedelta(meta_proto.execution_duration)
        return QueryMeta(
            execution_duration_s=execution_duration.total_seconds(),
            deployment_id=meta_proto.deployment_id,
            environment_id=meta_proto.environment_id,
            environment_name=meta_proto.environment_name,
            query_id=meta_proto.query_id,
            query_timestamp=meta_proto.query_timestamp.ToDatetime(tzinfo=dt.timezone.utc)
            if meta_proto.HasField("query_timestamp")
            else None,
            query_hash=meta_proto.query_hash,
            explain_output=get_field_or(meta_proto.explain_output, "plan_string"),
        )

    @staticmethod
    def online_query_request_decode(
        online_query_request: online_query_pb2.OnlineQueryRequest,
    ) -> OnlineQueryRequest:
        now_str: Optional[str] = None
        if online_query_request.HasField("now"):
            now_dt: dt.datetime = online_query_request.now.ToDatetime()
            now_str = now_dt.astimezone(tz=dt.timezone.utc).isoformat()
        return OnlineQueryRequest(
            inputs={k: proto_value_to_python(v) for k, v in online_query_request.inputs.items()},
            outputs=[output_expr.feature_fqn for output_expr in online_query_request.outputs],
            now=now_str,
            staleness={k: v for k, v in online_query_request.staleness.items()},
            context=OnlineQueryConverter._online_query_context_decode(online_query_request.context),
            include_meta=online_query_request.response_options.include_meta,
            explain=online_query_request.response_options.HasField("explain"),
            skip_online_storage=OnlineQueryConverter._options_bool_field(
                online_query_request.context, "skip_online_storage"
            ),
            skip_offline_storage=OnlineQueryConverter._options_bool_field(
                online_query_request.context, "skip_offline_storage"
            ),
            skip_metrics_storage=OnlineQueryConverter._options_bool_field(
                online_query_request.context, "skip_metrics_storage"
            ),
            correlation_id=online_query_request.context.correlation_id,
            query_name=online_query_request.context.query_name,
            query_name_version=online_query_request.context.query_name_version,
            deployment_id=online_query_request.context.deployment_id,
            branch_id=online_query_request.context.branch_id,
            meta={k: v for k, v in online_query_request.response_options.metadata.items()},
            store_plan_stages=OnlineQueryConverter._options_bool_field(
                online_query_request.context, "store_plan_stages"
            )
            or False,
            encoding_options=OnlineQueryConverter._feature_encoding_options_decode(
                online_query_request.response_options.encoding_options
            ),
            planner_options=None,  # for now
        )

    @staticmethod
    def online_query_context_encode(
        request: Union[OnlineQueryRequest, OnlineQueryManyRequest], options: Dict[str, Any]
    ):
        if request.context is None:
            raise ValueError("Given request object has no OnlineQueryContext.")
        context_options_proto = {k: value_to_proto(v) for k, v in options.items()}
        return online_query_pb2.OnlineQueryContext(
            environment=request.context.environment,
            tags=request.context.tags,
            required_resolver_tags=request.context.required_resolver_tags,
            deployment_id=request.deployment_id,
            branch_id=request.branch_id,
            correlation_id=request.correlation_id,
            query_name=request.query_name,
            query_name_version=request.query_name_version,
            options=context_options_proto,
        )

    @staticmethod
    def online_query_request_encode(request: OnlineQueryRequest) -> online_query_pb2.OnlineQueryRequest:
        inputs_proto = {k: value_to_proto(v) for k, v in request.inputs.items()}
        # TODO -- eventually we'll have more complex structured output expressions
        outputs_proto = [online_query_pb2.OutputExpr(feature_fqn=o) for o in request.outputs]
        now_proto = None
        if request.now is not None:
            now_proto = datetime_to_proto_timestamp(dt.datetime.fromisoformat(request.now))
        context_options_dict: Dict[str, Any] = {
            "skip_online_storage": request.skip_online_storage,
            "skip_offline_storage": request.skip_offline_storage,
            "skip_metrics_storage": request.skip_metrics_storage,
            "store_plan_stages": request.store_plan_stages,
        }
        context_options_dict.update(**(request.planner_options or {}))
        context_proto = OnlineQueryConverter.online_query_context_encode(request, options=context_options_dict)
        return online_query_pb2.OnlineQueryRequest(
            inputs=inputs_proto,
            outputs=outputs_proto,
            staleness=request.staleness,
            now=now_proto,
            context=context_proto,
            response_options=online_query_pb2.OnlineQueryResponseOptions(
                include_meta=request.include_meta,
                explain=online_query_pb2.ExplainOptions() if request.explain else None,
                encoding_options=online_query_pb2.FeatureEncodingOptions(
                    encode_structs_as_objects=request.encoding_options.encode_structs_as_objects
                ),
                metadata=request.meta,
            ),
        )

    @staticmethod
    def online_query_request_feather_encode(
        request: OnlineQueryManyRequest,
    ) -> online_query_pb2.OnlineQueryBulkRequest:
        import pyarrow as pa

        rb = pa.RecordBatch.from_pydict(request.inputs)
        inputs_bytes = record_batch_to_arrow_ipc(rb)
        outputs = [online_query_pb2.OutputExpr(feature_fqn=o) for o in request.outputs]
        now_proto = None
        if request.now is not None:
            now_proto = [datetime_to_proto_timestamp(dt.datetime.fromisoformat(n)) for n in request.now]
        context_options_dict = {
            "skip_online_storage": request.skip_online_storage,
            "skip_offline_storage": request.skip_offline_storage,
            "skip_metrics_storage": request.skip_metrics_storage,
            "store_plan_stages": request.store_plan_stages,
            "planner_version": "2",  # TODO remove this
        }
        context_proto = OnlineQueryConverter.online_query_context_encode(request, options=context_options_dict)
        return online_query_pb2.OnlineQueryBulkRequest(
            inputs_feather=inputs_bytes,
            outputs=outputs,
            staleness=request.staleness,
            now=now_proto,
            context=context_proto,
            response_options=online_query_pb2.OnlineQueryResponseOptions(
                include_meta=request.include_meta,
                explain=online_query_pb2.ExplainOptions() if request.explain else None,
                encoding_options=online_query_pb2.FeatureEncodingOptions(
                    encode_structs_as_objects=request.encoding_options.encode_structs_as_objects
                ),
                metadata=request.meta,
            ),
        )

    @staticmethod
    def _feature_meta_encode(feature_meta: FeatureResolutionMeta) -> online_query_pb2.FeatureMeta:
        return online_query_pb2.FeatureMeta(
            chosen_resolver_fqn=feature_meta.chosen_resolver_fqn,
            cache_hit=feature_meta.cache_hit,
            primitive_type=feature_meta.primitive_type,
            version=feature_meta.version,
        )

    @staticmethod
    def _feature_result_encode(feature_result: FeatureResult) -> online_query_pb2.FeatureResult:
        feature_result_proto = online_query_pb2.FeatureResult(
            field=feature_result.field,
        )
        feature_result_proto.value.CopyFrom(value_to_proto(feature_result.value))
        if feature_result.error:
            feature_result_proto.error.CopyFrom(ChalkErrorConverter.chalk_error_encode(feature_result.error))
        if feature_result.ts:
            feature_result_proto.ts.CopyFrom(datetime_to_proto_timestamp(feature_result.ts))
        if feature_result.meta:
            feature_result_proto.meta.CopyFrom(OnlineQueryConverter._feature_meta_encode(feature_result.meta))
        return feature_result_proto

    @staticmethod
    def _feature_result_decode(proto: online_query_pb2.FeatureResult) -> FeatureResult:
        return FeatureResult(
            field=proto.field,
            value=proto_value_to_python(proto.value),
            pkey=proto_value_to_python(proto.pkey),
            error=ChalkErrorConverter.chalk_error_decode(proto.error) if proto.HasField("error") else None,
            ts=proto.ts.ToDatetime(tzinfo=dt.timezone.utc) if proto.HasField("ts") else None,
            meta=FeatureResolutionMeta(
                chosen_resolver_fqn=proto.meta.chosen_resolver_fqn,
                cache_hit=proto.meta.cache_hit,
                primitive_type=proto.meta.primitive_type,
                version=proto.meta.version,
            )
            if proto.HasField("meta")
            else None,
        )

    @staticmethod
    def _online_query_result_decode(result_proto: online_query_pb2.OnlineQueryResult) -> List[FeatureResult]:
        results_py = []
        for fr_proto in result_proto.results:
            results_py.append(OnlineQueryConverter._feature_result_decode(fr_proto))
        return results_py

    @staticmethod
    def online_query_response_encode(
        online_query_result: OnlineQueryResponse,
    ) -> online_query_pb2.OnlineQueryResponse:
        proto_response = online_query_pb2.OnlineQueryResponse()
        proto_response.data.results.extend(
            [OnlineQueryConverter._feature_result_encode(fr) for fr in online_query_result.data]
        )
        if online_query_result.errors:
            proto_response.errors.extend(
                [ChalkErrorConverter.chalk_error_encode(e) for e in online_query_result.errors]
            )
        if online_query_result.meta:
            proto_response.response_meta.CopyFrom(OnlineQueryConverter._query_meta_encode(online_query_result.meta))
        return proto_response

    @staticmethod
    def online_query_response_decode(response_proto: online_query_pb2.OnlineQueryResponse) -> OnlineQueryResponse:
        data = []
        errors = []
        meta = None
        if response_proto.HasField("data"):
            data = OnlineQueryConverter._online_query_result_decode(response_proto.data)
        for err_proto in response_proto.errors:
            errors.append(ChalkErrorConverter.chalk_error_decode(err_proto))
        if response_proto.HasField("response_meta"):
            meta = OnlineQueryConverter._query_meta_decode(response_proto.response_meta)
        return OnlineQueryResponse(
            data=data,
            errors=errors,
            meta=meta,
        )

    @staticmethod
    def online_query_bulk_response_decode_to_single(
        response_proto: online_query_pb2.OnlineQueryBulkResponse,
    ) -> OnlineQueryResponse:
        import pyarrow as pa

        oat_prefix = ".__chalk_observed_at__"
        res: List[FeatureResult] = []

        if len(response_proto.scalars_data) > 0:
            root_ts: Optional[dt.datetime] = None
            pkey: Any = None
            batch: pa.RecordBatch = arrow_ipc_to_record_batch(response_proto.scalars_data)

            if len(batch) != 1:
                raise ValueError(f"Expected exactly one scalar data row in response, found {len(batch)}")

            for col_name in batch.schema.names:
                if col_name.endswith(oat_prefix):
                    root_ts = batch[col_name][0].as_py()
                    break
                if col_name == PKEY_COL_NAME:
                    pkey = batch[col_name][0].as_py()

            for col_name in batch.schema.names:
                if (
                    col_name.startswith("__chalk__.__metadata__.")
                    or col_name.startswith(GRPC_RESULT_METADATA_COL_PREFIX)
                    or col_name.endswith(oat_prefix)
                    or col_name in (TS_COL_NAME, INDEX_COL_NAME, PKEY_COL_NAME, OBSERVED_AT_COL_NAME)
                ):
                    continue

                col_field = batch.schema.field(col_name)
                converter = PrimitiveFeatureConverter(
                    name=col_name,
                    is_nullable=col_field.nullable,
                    pyarrow_dtype=col_field.type,
                )
                scoped_ts_fqn = (col_field.metadata or {}).get(b"ts_fqn")
                if scoped_ts_fqn is not None and scoped_ts_fqn in batch.schema.names:
                    scoped_ts_val = batch[scoped_ts_fqn][0].as_py()
                else:
                    scoped_ts_val = root_ts
                converted_col = converter.from_pyarrow_to_primitive(batch[col_name])
                assert (
                    len(converted_col) == 1
                ), f"Expected exactly one element in column {col_name}, found {len(converted_col)}"
                result = FeatureResult(field=col_name, value=converted_col[0], ts=scoped_ts_val, pkey=pkey)
                metadata_col_name = GRPC_RESULT_METADATA_COL_PREFIX + col_name
                if metadata_col_name in batch.schema.names:
                    metadata_col: pa.StructArray = batch[metadata_col_name]
                    result.valid = metadata_col[0].is_valid
                    if result.valid:
                        assert isinstance(metadata_col, pa.StructArray)
                        source_id_field = metadata_col.field("source_id")  # pyright: ignore
                        source_id: str | None = source_id_field[0].as_py()
                        source_type_field = metadata_col.field("source_type")  # pyright: ignore
                        source_type: str | None = source_type_field[0].as_py()
                        is_cache_hit = source_type in ["online_store", "offline_store"]
                        resolver_fqn = source_id or "unknown"
                        result.meta = FeatureResolutionMeta(
                            chosen_resolver_fqn=resolver_fqn,
                            cache_hit=is_cache_hit,
                            #    version=1
                        )

                res.append(result)

        for output_name, output_bytes in response_proto.groups_data.items():
            values = []
            namespace_ts: Optional[dt.datetime] = None
            batch = pa.Table.from_batches([arrow_ipc_to_record_batch(output_bytes)])

            for col_name in batch.schema.names:
                if col_name.endswith(oat_prefix):
                    namespace_ts = batch[col_name][0].as_py()
                    break

            for col_name in batch.schema.names:
                converter = PrimitiveFeatureConverter(
                    name=col_name,
                    is_nullable=batch.schema.field(col_name).nullable,
                    pyarrow_dtype=batch.schema.field(col_name).type,
                )
                values.append(converter.from_pyarrow_to_primitive(batch[col_name]))

            res.append(
                FeatureResult(
                    field=output_name,
                    value={
                        "columns": batch.schema.names,
                        "values": values,
                    },
                    ts=namespace_ts,
                )
            )

        return OnlineQueryResponse(
            data=res,
            meta=OnlineQueryConverter._query_meta_decode(response_proto.response_meta)
            if response_proto.HasField("response_meta")
            else None,
            errors=[ChalkErrorConverter.chalk_error_decode(e) for e in response_proto.errors],
        )

    @staticmethod
    def online_query_result_feather_encode(
        online_query_result_feather: OnlineQueryResultFeather,
    ) -> online_query_pb2.OnlineQueryBulkResponse:
        proto_response = online_query_pb2.OnlineQueryBulkResponse()
        proto_response.scalars_data = online_query_result_feather.scalar_data
        for k, v in online_query_result_feather.groups_data.items():
            # v should already be bytes at this point
            proto_response.groups_data[k] = v
        if online_query_result_feather.errors:
            for error in online_query_result_feather.errors:
                json_error = json.loads(error)
                chalk_error = ChalkError.create(**json_error)
                chalk_error_proto = chalk_error_pb2.ChalkError()
                chalk_error_proto.CopyFrom(ChalkErrorConverter.chalk_error_encode(chalk_error))
                proto_response.errors.append(chalk_error_proto)
        if online_query_result_feather.meta:
            meta_json = json.loads(online_query_result_feather.meta)
            query_meta = QueryMeta(**meta_json)
            proto_response.response_meta.CopyFrom(OnlineQueryConverter._query_meta_encode(query_meta))
        return proto_response

    @staticmethod
    def online_query_bulk_response_decode(
        response_proto: online_query_pb2.OnlineQueryBulkResponse, trace_id: Optional[str]
    ) -> BulkOnlineQueryResult:
        import polars as pl
        import pyarrow as pa

        scalars: Optional[pl.DataFrame] = None
        groups: Dict[str, pl.DataFrame] = {}
        errors = []
        meta = None
        if len(response_proto.scalars_data) > 0:
            scalars_batch = arrow_ipc_to_record_batch(response_proto.scalars_data)
            scalars = pa_table_to_pl_df(pa.Table.from_batches([scalars_batch]))
        for output_name, output_bytes in response_proto.groups_data.items():
            group_batch = arrow_ipc_to_record_batch(output_bytes)
            groups[output_name] = pa_table_to_pl_df(pa.Table.from_batches([group_batch]))
        for err_proto in response_proto.errors:
            errors.append(ChalkErrorConverter.chalk_error_decode(err_proto))
        if response_proto.HasField("response_meta"):
            meta = OnlineQueryConverter._query_meta_decode(response_proto.response_meta)
        return BulkOnlineQueryResult(scalars_df=scalars, groups_dfs=groups, meta=meta, errors=errors, trace_id=trace_id)

    @staticmethod
    def online_query_multi_response_decode(
        response_proto: online_query_pb2.OnlineQueryMultiResponse, trace_id: Optional[str]
    ) -> BulkOnlineQueryResponse:
        results = []
        for result_proto in response_proto.responses:
            if result_proto.HasField("single_response"):
                results.append(OnlineQueryConverter.online_query_response_decode(result_proto.single_response))
            elif result_proto.HasField("bulk_response"):
                # Add trace ID to sub-query result objects for clarity
                results.append(
                    OnlineQueryConverter.online_query_bulk_response_decode(
                        result_proto.bulk_response, trace_id=trace_id
                    )
                )
        global_errors = [ChalkErrorConverter.chalk_error_decode(e) for e in response_proto.errors]
        return BulkOnlineQueryResponse(results=results, global_errors=global_errors, trace_id=trace_id)



from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class AuditLevel(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    AUDIT_LEVEL_UNSPECIFIED: _ClassVar[AuditLevel]
    AUDIT_LEVEL_ERRORS: _ClassVar[AuditLevel]
    AUDIT_LEVEL_ALL: _ClassVar[AuditLevel]

AUDIT_LEVEL_UNSPECIFIED: AuditLevel
AUDIT_LEVEL_ERRORS: AuditLevel
AUDIT_LEVEL_ALL: AuditLevel
AUDIT_FIELD_NUMBER: _ClassVar[int]
audit: _descriptor.FieldDescriptor

class AuditOptions(_message.Message):
    __slots__ = ("level", "description")
    LEVEL_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    level: AuditLevel
    description: str
    def __init__(self, level: _Optional[_Union[AuditLevel, str]] = ..., description: _Optional[str] = ...) -> None: ...



from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class Webhook(_message.Message):
    __slots__ = (
        "id",
        "environment_id",
        "team_id",
        "name",
        "url",
        "subscriptions",
        "secret",
        "headers",
        "created_at",
        "updated_at",
    )
    ID_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    TEAM_ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    URL_FIELD_NUMBER: _ClassVar[int]
    SUBSCRIPTIONS_FIELD_NUMBER: _ClassVar[int]
    SECRET_FIELD_NUMBER: _ClassVar[int]
    HEADERS_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    id: str
    environment_id: str
    team_id: str
    name: str
    url: str
    subscriptions: _containers.RepeatedScalarFieldContainer[str]
    secret: str
    headers: _struct_pb2.Struct
    created_at: _timestamp_pb2.Timestamp
    updated_at: _timestamp_pb2.Timestamp
    def __init__(
        self,
        id: _Optional[str] = ...,
        environment_id: _Optional[str] = ...,
        team_id: _Optional[str] = ...,
        name: _Optional[str] = ...,
        url: _Optional[str] = ...,
        subscriptions: _Optional[_Iterable[str]] = ...,
        secret: _Optional[str] = ...,
        headers: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ...,
        created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
    ) -> None: ...

class CreateWebhookRequest(_message.Message):
    __slots__ = ("environment_id", "name", "url", "subscriptions", "secret", "headers")
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    URL_FIELD_NUMBER: _ClassVar[int]
    SUBSCRIPTIONS_FIELD_NUMBER: _ClassVar[int]
    SECRET_FIELD_NUMBER: _ClassVar[int]
    HEADERS_FIELD_NUMBER: _ClassVar[int]
    environment_id: str
    name: str
    url: str
    subscriptions: _containers.RepeatedScalarFieldContainer[str]
    secret: str
    headers: _struct_pb2.Struct
    def __init__(
        self,
        environment_id: _Optional[str] = ...,
        name: _Optional[str] = ...,
        url: _Optional[str] = ...,
        subscriptions: _Optional[_Iterable[str]] = ...,
        secret: _Optional[str] = ...,
        headers: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ...,
    ) -> None: ...

class CreateWebhookResponse(_message.Message):
    __slots__ = ("webhook",)
    WEBHOOK_FIELD_NUMBER: _ClassVar[int]
    webhook: Webhook
    def __init__(self, webhook: _Optional[_Union[Webhook, _Mapping]] = ...) -> None: ...

class UpdateWebhookRequest(_message.Message):
    __slots__ = ("id", "name", "url", "subscriptions", "secret", "headers")
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    URL_FIELD_NUMBER: _ClassVar[int]
    SUBSCRIPTIONS_FIELD_NUMBER: _ClassVar[int]
    SECRET_FIELD_NUMBER: _ClassVar[int]
    HEADERS_FIELD_NUMBER: _ClassVar[int]
    id: str
    name: str
    url: str
    subscriptions: _containers.RepeatedScalarFieldContainer[str]
    secret: str
    headers: _struct_pb2.Struct
    def __init__(
        self,
        id: _Optional[str] = ...,
        name: _Optional[str] = ...,
        url: _Optional[str] = ...,
        subscriptions: _Optional[_Iterable[str]] = ...,
        secret: _Optional[str] = ...,
        headers: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ...,
    ) -> None: ...

class UpdateWebhookResponse(_message.Message):
    __slots__ = ("webhook",)
    WEBHOOK_FIELD_NUMBER: _ClassVar[int]
    webhook: Webhook
    def __init__(self, webhook: _Optional[_Union[Webhook, _Mapping]] = ...) -> None: ...

class DeleteWebhookRequest(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class DeleteWebhookResponse(_message.Message):
    __slots__ = ("webhook",)
    WEBHOOK_FIELD_NUMBER: _ClassVar[int]
    webhook: Webhook
    def __init__(self, webhook: _Optional[_Union[Webhook, _Mapping]] = ...) -> None: ...

class GetWebhookRequest(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class GetWebhookResponse(_message.Message):
    __slots__ = ("webhook",)
    WEBHOOK_FIELD_NUMBER: _ClassVar[int]
    webhook: Webhook
    def __init__(self, webhook: _Optional[_Union[Webhook, _Mapping]] = ...) -> None: ...

class ListWebhooksRequest(_message.Message):
    __slots__ = ("environment_id",)
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    environment_id: str
    def __init__(self, environment_id: _Optional[str] = ...) -> None: ...

class ListWebhooksResponse(_message.Message):
    __slots__ = ("webhooks",)
    WEBHOOKS_FIELD_NUMBER: _ClassVar[int]
    webhooks: _containers.RepeatedCompositeFieldContainer[Webhook]
    def __init__(self, webhooks: _Optional[_Iterable[_Union[Webhook, _Mapping]]] = ...) -> None: ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.protosql.v1.sql_service_pb2 import (
    ExecuteSqlQueryRequest,
    ExecuteSqlQueryResponse,
    PlanSqlQueryRequest,
    PlanSqlQueryResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class SqlServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    ExecuteSqlQuery: UnaryUnaryMultiCallable[
        ExecuteSqlQueryRequest,
        ExecuteSqlQueryResponse,
    ]
    PlanSqlQuery: UnaryUnaryMultiCallable[
        PlanSqlQueryRequest,
        PlanSqlQueryResponse,
    ]

class SqlServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def ExecuteSqlQuery(
        self,
        request: ExecuteSqlQueryRequest,
        context: ServicerContext,
    ) -> ExecuteSqlQueryResponse: ...
    @abstractmethod
    def PlanSqlQuery(
        self,
        request: PlanSqlQueryRequest,
        context: ServicerContext,
    ) -> PlanSqlQueryResponse: ...

def add_SqlServiceServicer_to_server(servicer: SqlServiceServicer, server: Server) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/aggregate/v1/timeseries.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n#chalk/aggregate/v1/timeseries.proto\x12\x12\x63halk.aggregate.v1\x1a\x1egoogle/protobuf/duration.proto"\x92\x02\n\x17\x41ggregateTimeSeriesRule\x12 \n\x0b\x61ggregation\x18\x01 \x01(\tR\x0b\x61ggregation\x12\x42\n\x0f\x62ucket_duration\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationR\x0e\x62ucketDuration\x12-\n\x12\x64\x65pendent_features\x18\x03 \x03(\tR\x11\x64\x65pendentFeatures\x12\x37\n\tretention\x18\x04 \x01(\x0b\x32\x19.google.protobuf.DurationR\tretention\x12)\n\x10\x64\x61tetime_feature\x18\x05 \x01(\tR\x0f\x64\x61tetimeFeature"\x8c\x02\n\x13\x41ggregateTimeSeries\x12\x1c\n\tnamespace\x18\x01 \x01(\tR\tnamespace\x12!\n\x0c\x61ggregate_on\x18\x02 \x01(\tR\x0b\x61ggregateOn\x12\x19\n\x08group_by\x18\x03 \x03(\tR\x07groupBy\x12\x41\n\x05rules\x18\x05 \x03(\x0b\x32+.chalk.aggregate.v1.AggregateTimeSeriesRuleR\x05rules\x12/\n\x13\x66ilters_description\x18\x06 \x01(\tR\x12\x66iltersDescription\x12%\n\x0e\x62ucket_feature\x18\x07 \x01(\tR\rbucketFeatureB\xad\x01\n\x16\x63om.chalk.aggregate.v1B\x0fTimeseriesProtoP\x01Z\x18\x61ggregate/v1;aggregatev1\xa2\x02\x03\x43\x41X\xaa\x02\x12\x43halk.Aggregate.V1\xca\x02\x12\x43halk\\Aggregate\\V1\xe2\x02\x1e\x43halk\\Aggregate\\V1\\GPBMetadata\xea\x02\x14\x43halk::Aggregate::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.aggregate.v1.timeseries_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\026com.chalk.aggregate.v1B\017TimeseriesProtoP\001Z\030aggregate/v1;aggregatev1\242\002\003CAX\252\002\022Chalk.Aggregate.V1\312\002\022Chalk\\Aggregate\\V1\342\002\036Chalk\\Aggregate\\V1\\GPBMetadata\352\002\024Chalk::Aggregate::V1"
    _globals["_AGGREGATETIMESERIESRULE"]._serialized_start = 92
    _globals["_AGGREGATETIMESERIESRULE"]._serialized_end = 366
    _globals["_AGGREGATETIMESERIES"]._serialized_start = 369
    _globals["_AGGREGATETIMESERIES"]._serialized_end = 637
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/common/v1/dataset_response.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.common.v1 import chalk_error_pb2 as chalk_dot_common_dot_v1_dot_chalk__error__pb2
from chalk._gen.chalk.common.v1 import query_status_pb2 as chalk_dot_common_dot_v1_dot_query__status__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n&chalk/common/v1/dataset_response.proto\x12\x0f\x63halk.common.v1\x1a!chalk/common/v1/chalk_error.proto\x1a"chalk/common/v1/query_status.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\xef\x01\n\x13\x44\x61tasetSampleFilter\x12@\n\x0blower_bound\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x00R\nlowerBound\x88\x01\x01\x12@\n\x0bupper_bound\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x01R\nupperBound\x88\x01\x01\x12$\n\x0bmax_samples\x18\x03 \x01(\x05H\x02R\nmaxSamples\x88\x01\x01\x42\x0e\n\x0c_lower_boundB\x0e\n\x0c_upper_boundB\x0e\n\x0c_max_samples"\xa5\x01\n\rDatasetFilter\x12K\n\x0esample_filters\x18\x01 \x01(\x0b\x32$.chalk.common.v1.DatasetSampleFilterR\rsampleFilters\x12\x30\n\x12max_cache_age_secs\x18\x02 \x01(\x02H\x00R\x0fmaxCacheAgeSecs\x88\x01\x01\x42\x15\n\x13_max_cache_age_secs"\xcf\x07\n\x17\x44\x61tasetRevisionResponse\x12&\n\x0c\x64\x61taset_name\x18\x01 \x01(\tH\x00R\x0b\x64\x61tasetName\x88\x01\x01\x12"\n\ndataset_id\x18\x02 \x01(\tH\x01R\tdatasetId\x88\x01\x01\x12*\n\x0e\x65nvironment_id\x18\x03 \x01(\tH\x02R\renvironmentId\x88\x01\x01\x12$\n\x0brevision_id\x18\x04 \x01(\tH\x03R\nrevisionId\x88\x01\x01\x12\x1d\n\ncreator_id\x18\x05 \x01(\tR\tcreatorId\x12\x18\n\x07outputs\x18\x06 \x03(\tR\x07outputs\x12"\n\ngivens_uri\x18\x07 \x01(\tH\x04R\tgivensUri\x88\x01\x01\x12\x34\n\x06status\x18\x08 \x01(\x0e\x32\x1c.chalk.common.v1.QueryStatusR\x06status\x12\x38\n\x07\x66ilters\x18\t \x01(\x0b\x32\x1e.chalk.common.v1.DatasetFilterR\x07\x66ilters\x12%\n\x0enum_partitions\x18\n \x01(\x05R\rnumPartitions\x12 \n\tnum_bytes\x18\x0b \x01(\x05H\x05R\x08numBytes\x88\x01\x01\x12\x1f\n\x0boutput_uris\x18\x0c \x01(\tR\noutputUris\x12%\n\x0eoutput_version\x18\r \x01(\x05R\routputVersion\x12\x1b\n\x06\x62ranch\x18\x0e \x01(\tH\x06R\x06\x62ranch\x88\x01\x01\x12(\n\rdashboard_url\x18\x0f \x01(\tH\x07R\x0c\x64\x61shboardUrl\x88\x01\x01\x12>\n\ncreated_at\x18\x65 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x08R\tcreatedAt\x88\x01\x01\x12>\n\nstarted_at\x18\x66 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\tR\tstartedAt\x88\x01\x01\x12\x44\n\rterminated_at\x18g \x01(\x0b\x32\x1a.google.protobuf.TimestampH\nR\x0cterminatedAt\x88\x01\x01\x42\x0f\n\r_dataset_nameB\r\n\x0b_dataset_idB\x11\n\x0f_environment_idB\x0e\n\x0c_revision_idB\r\n\x0b_givens_uriB\x0c\n\n_num_bytesB\t\n\x07_branchB\x10\n\x0e_dashboard_urlB\r\n\x0b_created_atB\r\n\x0b_started_atB\x10\n\x0e_terminated_at"\xb2\x02\n\x0f\x44\x61tasetResponse\x12\x1f\n\x0bis_finished\x18\x01 \x01(\x08R\nisFinished\x12\x18\n\x07version\x18\x02 \x01(\x05R\x07version\x12%\n\x0e\x65nvironment_id\x18\x03 \x01(\tR\renvironmentId\x12\x1d\n\ndataset_id\x18\x04 \x01(\tR\tdatasetId\x12!\n\x0c\x64\x61taset_name\x18\x05 \x01(\tR\x0b\x64\x61tasetName\x12\x33\n\x06\x65rrors\x18\x65 \x03(\x0b\x32\x1b.chalk.common.v1.ChalkErrorR\x06\x65rrors\x12\x46\n\trevisions\x18\x66 \x03(\x0b\x32(.chalk.common.v1.DatasetRevisionResponseR\trevisionsB\x89\x01\n\x13\x63om.chalk.common.v1B\x14\x44\x61tasetResponseProtoP\x01\xa2\x02\x03\x43\x43X\xaa\x02\x0f\x43halk.Common.V1\xca\x02\x0f\x43halk\\Common\\V1\xe2\x02\x1b\x43halk\\Common\\V1\\GPBMetadata\xea\x02\x11\x43halk::Common::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.common.v1.dataset_response_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.common.v1B\024DatasetResponseProtoP\001\242\002\003CCX\252\002\017Chalk.Common.V1\312\002\017Chalk\\Common\\V1\342\002\033Chalk\\Common\\V1\\GPBMetadata\352\002\021Chalk::Common::V1"
    _globals["_DATASETSAMPLEFILTER"]._serialized_start = 164
    _globals["_DATASETSAMPLEFILTER"]._serialized_end = 403
    _globals["_DATASETFILTER"]._serialized_start = 406
    _globals["_DATASETFILTER"]._serialized_end = 571
    _globals["_DATASETREVISIONRESPONSE"]._serialized_start = 574
    _globals["_DATASETREVISIONRESPONSE"]._serialized_end = 1549
    _globals["_DATASETRESPONSE"]._serialized_start = 1552
    _globals["_DATASETRESPONSE"]._serialized_end = 1858
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import grpc

from chalk._gen.chalk.aggregate.v1 import service_pb2 as chalk_dot_aggregate_dot_v1_dot_service__pb2


class AggregateServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.PlanAggregateBackfill = channel.unary_unary(
            "/chalk.aggregate.v1.AggregateService/PlanAggregateBackfill",
            request_serializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.PlanAggregateBackfillRequest.SerializeToString,
            response_deserializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.PlanAggregateBackfillResponse.FromString,
        )
        self.GetAggregates = channel.unary_unary(
            "/chalk.aggregate.v1.AggregateService/GetAggregates",
            request_serializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetAggregatesRequest.SerializeToString,
            response_deserializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetAggregatesResponse.FromString,
        )
        self.GetAggregateBackfillJobs = channel.unary_unary(
            "/chalk.aggregate.v1.AggregateService/GetAggregateBackfillJobs",
            request_serializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetAggregateBackfillJobsRequest.SerializeToString,
            response_deserializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetAggregateBackfillJobsResponse.FromString,
        )
        self.GetAggregateBackfillJob = channel.unary_unary(
            "/chalk.aggregate.v1.AggregateService/GetAggregateBackfillJob",
            request_serializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetAggregateBackfillJobRequest.SerializeToString,
            response_deserializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetAggregateBackfillJobResponse.FromString,
        )
        self.GetCronAggregateBackfill = channel.unary_unary(
            "/chalk.aggregate.v1.AggregateService/GetCronAggregateBackfill",
            request_serializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetCronAggregateBackfillRequest.SerializeToString,
            response_deserializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetCronAggregateBackfillResponse.FromString,
        )
        self.GetActiveCronAggregateBackfills = channel.unary_unary(
            "/chalk.aggregate.v1.AggregateService/GetActiveCronAggregateBackfills",
            request_serializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetActiveCronAggregateBackfillsRequest.SerializeToString,
            response_deserializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetActiveCronAggregateBackfillsResponse.FromString,
        )


class AggregateServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def PlanAggregateBackfill(self, request, context):
        """PlanAggregateBackfill determines the estimated resources needed to backfill
        an aggregate.

        This method is a duplicate of the PlanAggregateBackfill method
        in the query_server.proto file. We should remove the query_server.proto method
        and move that request to this service instead.
        buf:lint:ignore RPC_REQUEST_RESPONSE_UNIQUE
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetAggregates(self, request, context):
        """This method is a duplicate of the PlanAggregateBackfill method
        in the query_server.proto file. We should remove the query_server.proto method
        and move that request to this service instead.
        buf:lint:ignore RPC_REQUEST_RESPONSE_UNIQUE
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetAggregateBackfillJobs(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetAggregateBackfillJob(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetCronAggregateBackfill(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetActiveCronAggregateBackfills(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_AggregateServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "PlanAggregateBackfill": grpc.unary_unary_rpc_method_handler(
            servicer.PlanAggregateBackfill,
            request_deserializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.PlanAggregateBackfillRequest.FromString,
            response_serializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.PlanAggregateBackfillResponse.SerializeToString,
        ),
        "GetAggregates": grpc.unary_unary_rpc_method_handler(
            servicer.GetAggregates,
            request_deserializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetAggregatesRequest.FromString,
            response_serializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetAggregatesResponse.SerializeToString,
        ),
        "GetAggregateBackfillJobs": grpc.unary_unary_rpc_method_handler(
            servicer.GetAggregateBackfillJobs,
            request_deserializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetAggregateBackfillJobsRequest.FromString,
            response_serializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetAggregateBackfillJobsResponse.SerializeToString,
        ),
        "GetAggregateBackfillJob": grpc.unary_unary_rpc_method_handler(
            servicer.GetAggregateBackfillJob,
            request_deserializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetAggregateBackfillJobRequest.FromString,
            response_serializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetAggregateBackfillJobResponse.SerializeToString,
        ),
        "GetCronAggregateBackfill": grpc.unary_unary_rpc_method_handler(
            servicer.GetCronAggregateBackfill,
            request_deserializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetCronAggregateBackfillRequest.FromString,
            response_serializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetCronAggregateBackfillResponse.SerializeToString,
        ),
        "GetActiveCronAggregateBackfills": grpc.unary_unary_rpc_method_handler(
            servicer.GetActiveCronAggregateBackfills,
            request_deserializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetActiveCronAggregateBackfillsRequest.FromString,
            response_serializer=chalk_dot_aggregate_dot_v1_dot_service__pb2.GetActiveCronAggregateBackfillsResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler("chalk.aggregate.v1.AggregateService", rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class AggregateService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def PlanAggregateBackfill(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.aggregate.v1.AggregateService/PlanAggregateBackfill",
            chalk_dot_aggregate_dot_v1_dot_service__pb2.PlanAggregateBackfillRequest.SerializeToString,
            chalk_dot_aggregate_dot_v1_dot_service__pb2.PlanAggregateBackfillResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetAggregates(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.aggregate.v1.AggregateService/GetAggregates",
            chalk_dot_aggregate_dot_v1_dot_service__pb2.GetAggregatesRequest.SerializeToString,
            chalk_dot_aggregate_dot_v1_dot_service__pb2.GetAggregatesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetAggregateBackfillJobs(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.aggregate.v1.AggregateService/GetAggregateBackfillJobs",
            chalk_dot_aggregate_dot_v1_dot_service__pb2.GetAggregateBackfillJobsRequest.SerializeToString,
            chalk_dot_aggregate_dot_v1_dot_service__pb2.GetAggregateBackfillJobsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetAggregateBackfillJob(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.aggregate.v1.AggregateService/GetAggregateBackfillJob",
            chalk_dot_aggregate_dot_v1_dot_service__pb2.GetAggregateBackfillJobRequest.SerializeToString,
            chalk_dot_aggregate_dot_v1_dot_service__pb2.GetAggregateBackfillJobResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetCronAggregateBackfill(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.aggregate.v1.AggregateService/GetCronAggregateBackfill",
            chalk_dot_aggregate_dot_v1_dot_service__pb2.GetCronAggregateBackfillRequest.SerializeToString,
            chalk_dot_aggregate_dot_v1_dot_service__pb2.GetCronAggregateBackfillResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetActiveCronAggregateBackfills(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.aggregate.v1.AggregateService/GetActiveCronAggregateBackfills",
            chalk_dot_aggregate_dot_v1_dot_service__pb2.GetActiveCronAggregateBackfillsRequest.SerializeToString,
            chalk_dot_aggregate_dot_v1_dot_service__pb2.GetActiveCronAggregateBackfillsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



from chalk._gen.chalk.common.v1 import chalk_error_pb2 as _chalk_error_pb2
from chalk._gen.chalk.common.v1 import online_query_pb2 as _online_query_pb2
from chalk._gen.chalk.expression.v1 import expression_pb2 as _expression_pb2
from chalk._gen.chalk.graph.v1 import graph_pb2 as _graph_pb2
from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class OfflineQueryRecomputeFeatures(_message.Message):
    __slots__ = ("all_or_none", "feature_list")
    class FeatureList(_message.Message):
        __slots__ = ("feature_list",)
        FEATURE_LIST_FIELD_NUMBER: _ClassVar[int]
        feature_list: _containers.RepeatedScalarFieldContainer[str]
        def __init__(self, feature_list: _Optional[_Iterable[str]] = ...) -> None: ...

    ALL_OR_NONE_FIELD_NUMBER: _ClassVar[int]
    FEATURE_LIST_FIELD_NUMBER: _ClassVar[int]
    all_or_none: bool
    feature_list: OfflineQueryRecomputeFeatures.FeatureList
    def __init__(
        self,
        all_or_none: bool = ...,
        feature_list: _Optional[_Union[OfflineQueryRecomputeFeatures.FeatureList, _Mapping]] = ...,
    ) -> None: ...

class OfflineQueryExplain(_message.Message):
    __slots__ = ("truthy", "only")
    class Only(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...

    TRUTHY_FIELD_NUMBER: _ClassVar[int]
    ONLY_FIELD_NUMBER: _ClassVar[int]
    truthy: bool
    only: OfflineQueryExplain.Only
    def __init__(
        self, truthy: bool = ..., only: _Optional[_Union[OfflineQueryExplain.Only, _Mapping]] = ...
    ) -> None: ...

class OfflineQueryInput(_message.Message):
    __slots__ = ("columns", "values")
    COLUMNS_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    columns: _containers.RepeatedScalarFieldContainer[str]
    values: _containers.RepeatedCompositeFieldContainer[_struct_pb2.ListValue]
    def __init__(
        self,
        columns: _Optional[_Iterable[str]] = ...,
        values: _Optional[_Iterable[_Union[_struct_pb2.ListValue, _Mapping]]] = ...,
    ) -> None: ...

class OfflineQueryInputSharded(_message.Message):
    __slots__ = ("inputs",)
    INPUTS_FIELD_NUMBER: _ClassVar[int]
    inputs: _containers.RepeatedCompositeFieldContainer[OfflineQueryInput]
    def __init__(self, inputs: _Optional[_Iterable[_Union[OfflineQueryInput, _Mapping]]] = ...) -> None: ...

class OfflineQueryInputs(_message.Message):
    __slots__ = ("feather_inputs", "no_inputs", "single_inputs", "sharded_inputs")
    class NoInputs(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...

    FEATHER_INPUTS_FIELD_NUMBER: _ClassVar[int]
    NO_INPUTS_FIELD_NUMBER: _ClassVar[int]
    SINGLE_INPUTS_FIELD_NUMBER: _ClassVar[int]
    SHARDED_INPUTS_FIELD_NUMBER: _ClassVar[int]
    feather_inputs: bytes
    no_inputs: OfflineQueryInputs.NoInputs
    single_inputs: OfflineQueryInput
    sharded_inputs: OfflineQueryInputSharded
    def __init__(
        self,
        feather_inputs: _Optional[bytes] = ...,
        no_inputs: _Optional[_Union[OfflineQueryInputs.NoInputs, _Mapping]] = ...,
        single_inputs: _Optional[_Union[OfflineQueryInput, _Mapping]] = ...,
        sharded_inputs: _Optional[_Union[OfflineQueryInputSharded, _Mapping]] = ...,
    ) -> None: ...

class OfflineQueryRequest(_message.Message):
    __slots__ = (
        "inputs",
        "outputs",
        "required_outputs",
        "destination_format",
        "branch",
        "dataset_name",
        "recompute_features",
        "store_plan_stages",
        "filters",
        "spine_sql_query",
        "max_samples",
        "max_cache_age_secs",
        "explain",
        "explain2",
        "tags",
        "correlation_id",
        "required_resolver_tags",
        "observed_at_lower_bound",
        "observed_at_upper_bound",
        "planner_options",
        "store_online",
        "store_offline",
        "use_multiple_computers",
        "num_shards",
        "num_workers",
        "query_context",
        "overlay_graph",
    )
    class PlannerOptionsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _struct_pb2.Value
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[_Union[_struct_pb2.Value, _Mapping]] = ...
        ) -> None: ...

    class QueryContextEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _struct_pb2.Value
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[_Union[_struct_pb2.Value, _Mapping]] = ...
        ) -> None: ...

    INPUTS_FIELD_NUMBER: _ClassVar[int]
    OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    REQUIRED_OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    DESTINATION_FORMAT_FIELD_NUMBER: _ClassVar[int]
    BRANCH_FIELD_NUMBER: _ClassVar[int]
    DATASET_NAME_FIELD_NUMBER: _ClassVar[int]
    RECOMPUTE_FEATURES_FIELD_NUMBER: _ClassVar[int]
    STORE_PLAN_STAGES_FIELD_NUMBER: _ClassVar[int]
    FILTERS_FIELD_NUMBER: _ClassVar[int]
    SPINE_SQL_QUERY_FIELD_NUMBER: _ClassVar[int]
    MAX_SAMPLES_FIELD_NUMBER: _ClassVar[int]
    MAX_CACHE_AGE_SECS_FIELD_NUMBER: _ClassVar[int]
    EXPLAIN_FIELD_NUMBER: _ClassVar[int]
    EXPLAIN2_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    CORRELATION_ID_FIELD_NUMBER: _ClassVar[int]
    REQUIRED_RESOLVER_TAGS_FIELD_NUMBER: _ClassVar[int]
    OBSERVED_AT_LOWER_BOUND_FIELD_NUMBER: _ClassVar[int]
    OBSERVED_AT_UPPER_BOUND_FIELD_NUMBER: _ClassVar[int]
    PLANNER_OPTIONS_FIELD_NUMBER: _ClassVar[int]
    STORE_ONLINE_FIELD_NUMBER: _ClassVar[int]
    STORE_OFFLINE_FIELD_NUMBER: _ClassVar[int]
    USE_MULTIPLE_COMPUTERS_FIELD_NUMBER: _ClassVar[int]
    NUM_SHARDS_FIELD_NUMBER: _ClassVar[int]
    NUM_WORKERS_FIELD_NUMBER: _ClassVar[int]
    QUERY_CONTEXT_FIELD_NUMBER: _ClassVar[int]
    OVERLAY_GRAPH_FIELD_NUMBER: _ClassVar[int]
    inputs: OfflineQueryInputs
    outputs: _containers.RepeatedScalarFieldContainer[str]
    required_outputs: _containers.RepeatedScalarFieldContainer[str]
    destination_format: str
    branch: str
    dataset_name: str
    recompute_features: OfflineQueryRecomputeFeatures
    store_plan_stages: bool
    filters: _containers.RepeatedCompositeFieldContainer[_expression_pb2.LogicalExprNode]
    spine_sql_query: str
    max_samples: int
    max_cache_age_secs: int
    explain: OfflineQueryExplain
    explain2: _online_query_pb2.ExplainOptions
    tags: _containers.RepeatedScalarFieldContainer[str]
    correlation_id: str
    required_resolver_tags: _containers.RepeatedScalarFieldContainer[str]
    observed_at_lower_bound: str
    observed_at_upper_bound: str
    planner_options: _containers.MessageMap[str, _struct_pb2.Value]
    store_online: bool
    store_offline: bool
    use_multiple_computers: bool
    num_shards: int
    num_workers: int
    query_context: _containers.MessageMap[str, _struct_pb2.Value]
    overlay_graph: _graph_pb2.OverlayGraph
    def __init__(
        self,
        inputs: _Optional[_Union[OfflineQueryInputs, _Mapping]] = ...,
        outputs: _Optional[_Iterable[str]] = ...,
        required_outputs: _Optional[_Iterable[str]] = ...,
        destination_format: _Optional[str] = ...,
        branch: _Optional[str] = ...,
        dataset_name: _Optional[str] = ...,
        recompute_features: _Optional[_Union[OfflineQueryRecomputeFeatures, _Mapping]] = ...,
        store_plan_stages: bool = ...,
        filters: _Optional[_Iterable[_Union[_expression_pb2.LogicalExprNode, _Mapping]]] = ...,
        spine_sql_query: _Optional[str] = ...,
        max_samples: _Optional[int] = ...,
        max_cache_age_secs: _Optional[int] = ...,
        explain: _Optional[_Union[OfflineQueryExplain, _Mapping]] = ...,
        explain2: _Optional[_Union[_online_query_pb2.ExplainOptions, _Mapping]] = ...,
        tags: _Optional[_Iterable[str]] = ...,
        correlation_id: _Optional[str] = ...,
        required_resolver_tags: _Optional[_Iterable[str]] = ...,
        observed_at_lower_bound: _Optional[str] = ...,
        observed_at_upper_bound: _Optional[str] = ...,
        planner_options: _Optional[_Mapping[str, _struct_pb2.Value]] = ...,
        store_online: bool = ...,
        store_offline: bool = ...,
        use_multiple_computers: bool = ...,
        num_shards: _Optional[int] = ...,
        num_workers: _Optional[int] = ...,
        query_context: _Optional[_Mapping[str, _struct_pb2.Value]] = ...,
        overlay_graph: _Optional[_Union[_graph_pb2.OverlayGraph, _Mapping]] = ...,
    ) -> None: ...

class ColumnMetadataList(_message.Message):
    __slots__ = ("metadata",)
    class ColumnMetadata(_message.Message):
        __slots__ = ("feature_fqn", "column_name", "dtype")
        FEATURE_FQN_FIELD_NUMBER: _ClassVar[int]
        COLUMN_NAME_FIELD_NUMBER: _ClassVar[int]
        DTYPE_FIELD_NUMBER: _ClassVar[int]
        feature_fqn: str
        column_name: str
        dtype: str
        def __init__(
            self, feature_fqn: _Optional[str] = ..., column_name: _Optional[str] = ..., dtype: _Optional[str] = ...
        ) -> None: ...

    METADATA_FIELD_NUMBER: _ClassVar[int]
    metadata: _containers.RepeatedCompositeFieldContainer[ColumnMetadataList.ColumnMetadata]
    def __init__(
        self, metadata: _Optional[_Iterable[_Union[ColumnMetadataList.ColumnMetadata, _Mapping]]] = ...
    ) -> None: ...

class GetOfflineQueryJobResponse(_message.Message):
    __slots__ = ("is_finished", "version", "urls", "errors", "columns")
    IS_FINISHED_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    URLS_FIELD_NUMBER: _ClassVar[int]
    ERRORS_FIELD_NUMBER: _ClassVar[int]
    COLUMNS_FIELD_NUMBER: _ClassVar[int]
    is_finished: bool
    version: int
    urls: _containers.RepeatedScalarFieldContainer[str]
    errors: _containers.RepeatedCompositeFieldContainer[_chalk_error_pb2.ChalkError]
    columns: ColumnMetadataList
    def __init__(
        self,
        is_finished: bool = ...,
        version: _Optional[int] = ...,
        urls: _Optional[_Iterable[str]] = ...,
        errors: _Optional[_Iterable[_Union[_chalk_error_pb2.ChalkError, _Mapping]]] = ...,
        columns: _Optional[_Union[ColumnMetadataList, _Mapping]] = ...,
    ) -> None: ...



from __future__ import annotations

import base64
import concurrent.futures
import contextlib
import dataclasses
import functools
import json
import os
import queue
import typing
import uuid
from typing import TYPE_CHECKING, Any, Callable, Mapping, NewType, Optional, Sequence, cast

import orjson
import packaging.version
import pyarrow
import pyarrow as pa
import pyarrow.compute as pc

from chalk.clogging import chalk_logger
from chalk.features import Feature
from chalk.features._encoding.converter import FeatureConverter
from chalk.integrations.named import create_integration_variable, load_integration_variable
from chalk.sql._internal.query_execution_parameters import QueryExecutionParameters
from chalk.sql._internal.sql_source import BaseSQLSource, SQLSourceKind, validate_dtypes_for_efficient_execution
from chalk.sql.finalized_query import FinalizedChalkQuery
from chalk.utils.df_utils import is_list_like, pa_array_to_pl_series
from chalk.utils.missing_dependency import missing_dependency_exception
from chalk.utils.threading import DEFAULT_IO_EXECUTOR, MultiSemaphore
from chalk.utils.tracing import safe_incr, safe_set_gauge

if TYPE_CHECKING:
    from snowflake.connector.connection import SnowflakeConnection
    from snowflake.connector.result_batch import ResultBatch
    from sqlalchemy.engine import Connection
    from sqlalchemy.engine.url import URL
    from sqlalchemy.sql.ddl import CreateTable, DropTable
    from sqlalchemy.sql.schema import Table

try:
    import sqlalchemy as sa
except ImportError:
    sa = None

if sa is None:
    _supported_sqlalchemy_types_for_pa_querying = ()
else:
    _supported_sqlalchemy_types_for_pa_querying = (
        sa.BigInteger,
        sa.Boolean,
        sa.BINARY,
        sa.BLOB,
        sa.LargeBinary,
        sa.Float,
        sa.Integer,
        sa.Time,
        sa.String,
        sa.Text,
        sa.VARBINARY,
        sa.DateTime,
        sa.Date,
        sa.SmallInteger,
        sa.BIGINT,
        sa.BOOLEAN,
        sa.CHAR,
        sa.DATETIME,
        sa.FLOAT,
        sa.INTEGER,
        sa.SMALLINT,
        sa.TEXT,
        sa.TIMESTAMP,
        sa.VARCHAR,
    )


@functools.lru_cache(None)
def _has_new_fetch_arrow_all():
    # The api for fetch arrow all changed in v3.7.0 to include force_return_table
    # See https://github.com/snowflakedb/snowflake-connector-python/blob/3cced62f2d31b84299b544222c836a275a6d45a2/src/snowflake/connector/cursor.py#L1344
    import snowflake.connector

    return packaging.version.parse(snowflake.connector.__version__) >= packaging.version.parse("3.7.0")


_WorkerId = NewType("_WorkerId", int)


def _stage_prefix(unload_job_identifier: str, snowflake_unload_stage: str) -> str:
    return f"{snowflake_unload_stage}/{unload_job_identifier}/"


def _rewrite_query_for_unload(sql: str, unload_job_identifier: str, snowflake_unload_stage: str):
    rewritten_sql = sql.rstrip(";\n ")
    prefix = _stage_prefix(unload_job_identifier, snowflake_unload_stage)
    new_query = f"""
    COPY INTO {prefix} FROM ({rewritten_sql}) file_format = (type = 'parquet') overwrite=true header=true; /* {json.dumps({'unload_job_identifier': unload_job_identifier})} */
    """
    return new_query


class ResultHandle(typing.Protocol):
    @property
    def estimated_uncompressed_size(self) -> int | None:
        return None

    def to_arrow(self, connection: SnowflakeConnection | None) -> pyarrow.Table:
        ...


@dataclasses.dataclass(frozen=True)
class ResultBatchResultHandle(ResultHandle):
    result_batch: ResultBatch

    @property
    def estimated_uncompressed_size(self) -> int | None:
        return self.result_batch.uncompressed_size

    @property
    def num_rows(self) -> int | None:
        return self.result_batch.rowcount

    def to_arrow(self, connection: SnowflakeConnection | None):
        from snowflake.connector.result_batch import ArrowResultBatch

        if isinstance(self.result_batch, ArrowResultBatch):
            return self.result_batch.to_arrow(connection)
        else:
            return self.result_batch.to_arrow()


@dataclasses.dataclass(frozen=True)
class UnloadedStorageFileResultHandle(ResultHandle):
    uri: str
    compressed_size: int

    @property
    def estimated_uncompressed_size(self) -> int | None:
        # this is a (hopefully) conservative estimate.
        return self.compressed_size * 8

    def to_arrow(self, connection: SnowflakeConnection | None):
        import polars as pl

        chalk_logger.info(f"Loading table from {self.uri}")
        return pl.read_parquet(source=self.uri).to_arrow()


_SNOWFLAKE_ACCOUNT_ID_NAME = "SNOWFLAKE_ACCOUNT_ID"
_SNOWFLAKE_WAREHOUSE_NAME = "SNOWFLAKE_WAREHOUSE"
_SNOWFLAKE_USER_NAME = "SNOWFLAKE_USER"
_SNOWFLAKE_PASSWORD_NAME = "SNOWFLAKE_PASSWORD"
_SNOWFLAKE_DATABASE_NAME = "SNOWFLAKE_DATABASE"
_SNOWFLAKE_SCHEMA_NAME = "SNOWFLAKE_SCHEMA"
_SNOWFLAKE_ROLE_NAME = "SNOWFLAKE_ROLE"
_SNOWFLAKE_PRIVATE_KEY_B64_NAME = "SNOWFLAKE_PRIVATE_KEY_B64"


class SnowflakeSourceImpl(BaseSQLSource):
    def __init__(
        self,
        *,
        name: Optional[str] = None,
        account_identifier: Optional[str] = None,
        warehouse: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        db: Optional[str] = None,
        schema: Optional[str] = None,
        role: Optional[str] = None,
        private_key_b64: Optional[str] = None,
        engine_args: Optional[dict[str, Any]] = None,
        executor: Optional[concurrent.futures.ThreadPoolExecutor] = None,
        integration_variable_override: Optional[Mapping[str, str]] = None,
    ):
        try:
            import snowflake.connector  # noqa
            import snowflake.sqlalchemy  # noqa
            from cryptography.hazmat.backends import default_backend
            from cryptography.hazmat.primitives import serialization
        except ModuleNotFoundError:
            raise missing_dependency_exception("chalkpy[snowflake]")
        del snowflake  # unused

        self.account_identifier = account_identifier or load_integration_variable(
            integration_name=name, name=_SNOWFLAKE_ACCOUNT_ID_NAME, override=integration_variable_override
        )
        self.warehouse = warehouse or load_integration_variable(
            integration_name=name, name=_SNOWFLAKE_WAREHOUSE_NAME, override=integration_variable_override
        )
        self.user = user or load_integration_variable(
            integration_name=name, name=_SNOWFLAKE_USER_NAME, override=integration_variable_override
        )
        self.password = password or load_integration_variable(
            integration_name=name, name=_SNOWFLAKE_PASSWORD_NAME, override=integration_variable_override
        )
        self.db = db or load_integration_variable(
            integration_name=name, name=_SNOWFLAKE_DATABASE_NAME, override=integration_variable_override
        )
        self.schema = schema or load_integration_variable(
            integration_name=name, name=_SNOWFLAKE_SCHEMA_NAME, override=integration_variable_override
        )
        self.role = role or load_integration_variable(
            integration_name=name, name=_SNOWFLAKE_ROLE_NAME, override=integration_variable_override
        )
        private_key_b64_str = private_key_b64
        if isinstance(private_key_b64_str, bytes):
            private_key_b64_str = private_key_b64_str.decode("utf-8")
        self.private_key_b64: Optional[str] = private_key_b64_str or load_integration_variable(
            integration_name=name, name=_SNOWFLAKE_PRIVATE_KEY_B64_NAME, override=integration_variable_override
        )
        self.executor = executor or DEFAULT_IO_EXECUTOR

        if engine_args is None:
            engine_args = {}
        connect_args = {
            "client_prefetch_threads": min((os.cpu_count() or 1) * 2, 32),
            "client_session_keep_alive": True,
            "application_name": "chalkai_featurepipelines",
            "application": "chalkai_featurepipelines",
        }
        if self.private_key_b64 is not None:
            raw_bytes = base64.b64decode(self.private_key_b64)
            # From https://docs.snowflake.com/en/developer-guide/python-connector/python-connector-connect#label-python-key-pair-authn-rotation
            private_key_bytes = serialization.load_pem_private_key(
                raw_bytes, password=None, backend=default_backend()
            ).private_bytes(
                encoding=serialization.Encoding.DER,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )
            connect_args = connect_args | {"private_key": private_key_bytes}
        engine_args.setdefault("pool_size", 20)
        engine_args.setdefault("max_overflow", 60)
        engine_args.setdefault("connect_args", connect_args)

        BaseSQLSource.__init__(self, name=name, engine_args=engine_args, async_engine_args={})

    kind = SQLSourceKind.snowflake

    def get_sqlglot_dialect(self) -> str | None:
        return "snowflake"

    def local_engine_url(self) -> URL:
        from sqlalchemy.engine.url import URL

        query = {
            k: v
            for k, v in (
                {
                    "database": self.db,
                    "schema": self.schema,
                    "warehouse": self.warehouse,
                    "role": self.role,
                }
            ).items()
            if v is not None
        }
        return URL.create(
            drivername="snowflake",
            username=self.user,
            password=self.password,
            host=self.account_identifier,
            query=query,
        )

    def convert_db_types(self, v: Any, converter: FeatureConverter):
        """
        Overload this if a given DB type needs custom type conversion
        """
        if (is_list_like(converter.pyarrow_dtype) or pa.types.is_struct(converter.pyarrow_dtype)) and isinstance(
            v, (str, bytes)
        ):
            # Need to json-decode these types
            v = orjson.loads(v)
        return converter.from_rich_to_primitive(v, missing_value_strategy="default_or_allow")

    @contextlib.contextmanager
    def _create_temp_table(
        self,
        create_temp_table: CreateTable,
        temp_table: Table,
        drop_temp_table: DropTable,
        connection: Connection,
        temp_value: pa.Table,
    ):
        from snowflake.connector import pandas_tools

        snowflake_cnx = cast("SnowflakeConnection", connection.connection.dbapi_connection)
        with snowflake_cnx.cursor() as cursor:
            chalk_logger.info(f"Creating temporary table {temp_table.name} in Snowflake.")
            cursor.execute(create_temp_table.compile(dialect=self.get_sqlalchemy_dialect()).string)
            try:
                pandas_tools.write_pandas(
                    cursor.connection,
                    temp_value.to_pandas(),
                    str(temp_table.name),
                )
                yield
            finally:
                # "temp table", to snowflake, means that it belongs to the session. However, we keep using the same Snowflake session
                chalk_logger.info(f"Dropping temporary table {temp_table.name} in Snowflake.")
                cursor.execute(drop_temp_table.compile(dialect=self.get_sqlalchemy_dialect()).string)

    def _postprocess_table(self, features: Mapping[str, Feature], tbl: pa.Table):
        columns: list[pa.Array] = []
        column_names: list[str] = []
        chalk_logger.info(
            f"Received a PyArrow table from Snowflake with {len(tbl)} rows; {len(tbl.column_names)} columns; {tbl.nbytes=}; {tbl.schema=}"
        )

        for col_name, feature in features.items():
            try:
                column = tbl[col_name]
                expected_type = feature.converter.pyarrow_dtype
                actual_type = tbl.schema.field(col_name).type
                if pa.types.is_list(expected_type) or pa.types.is_large_list(expected_type):
                    if pa.types.is_string(actual_type) or pa.types.is_large_string(actual_type):
                        series = pa_array_to_pl_series(tbl[col_name])
                        column = series.str.json_extract(feature.converter.polars_dtype).to_arrow().cast(expected_type)
                if actual_type != expected_type:
                    column = column.cast(options=pc.CastOptions(target_type=expected_type, allow_time_truncate=True))
                if isinstance(column, pa.ChunkedArray):
                    column = column.combine_chunks()
                columns.append(column)
                column_names.append(feature.root_fqn)
            except:
                chalk_logger.error(f"Failed to deserialize column '{col_name}' into '{feature}'", exc_info=True)
                raise

        return pa.RecordBatch.from_arrays(arrays=columns, names=column_names)

    def _download_worker(
        self,
        result_handles: queue.Queue[ResultHandle],
        sem: MultiSemaphore | None,
        pa_table_queue: queue.Queue[tuple[pa.Table, int] | _WorkerId],
        worker_idx: _WorkerId,
    ):
        try:
            while True:
                try:
                    x = result_handles.get_nowait()
                except queue.Empty:
                    break
                weight = x.estimated_uncompressed_size
                as_arrow = None
                if weight is None:
                    # This is possible if the chunk is "local", which I think snowflake does for small result batches
                    as_arrow: pyarrow.Table = x.to_arrow(None)
                    weight = as_arrow.nbytes
                if sem:
                    if weight > sem.initial_value:
                        # If the file is larger than the maximum size, we'll truncate it, so this file will be the only one being downloaded
                        weight = sem.initial_value
                    if weight > 0:
                        # No need to acquire the semaphore for empty tables
                        if not sem.acquire(weight):
                            raise RuntimeError("Failed to acquire semaphore for snowflake download")
                        safe_set_gauge("chalk.snowflake.remaining_prefetch_bytes", sem.get_value())
                if as_arrow is None:
                    with self.get_engine().connect() as connection:
                        # Reusing an existing connection from the pool to avoid re-establishing sessions, tls handshakes, etc...
                        snowflake_cnx = cast(
                            "SnowflakeConnection",
                            connection.connection.dbapi_connection,
                        )
                        as_arrow = x.to_arrow(snowflake_cnx)
                pa_table_queue.put((as_arrow, weight))
        finally:
            # At the end, putting the worker id to signal that this worker is done
            pa_table_queue.put(worker_idx)

    def supports_inefficient_fallback(self) -> bool:
        return False

    def _execute_query_efficient(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
        connection: Optional[Connection],
        query_execution_parameters: QueryExecutionParameters,
    ):
        # these imports are safe because the only way we end up here is if we have a valid SnowflakeSource constructed,
        # which already gates this import
        import snowflake.connector
        from sqlalchemy.sql import Select

        if isinstance(finalized_query.query, Select):
            validate_dtypes_for_efficient_execution(finalized_query.query, _supported_sqlalchemy_types_for_pa_querying)

        result_handles: queue.Queue[ResultHandle] = queue.Queue()  # Using a queue since we'll be concurrently reading
        with (
            self.get_engine().connect() if connection is None else contextlib.nullcontext(connection)
        ) as sqlalchemy_cnx:
            con = cast(
                snowflake.connector.SnowflakeConnection,
                sqlalchemy_cnx.connection.dbapi_connection,
            )
            chalk_logger.info("Established connection with Snowflake")
            sql, positional_params, named_params = self.compile_query(finalized_query)
            assert len(positional_params) == 0, "using named param style"
            with contextlib.ExitStack() as exit_stack:
                for (
                    _,
                    temp_value,
                    create_temp_table,
                    temp_table,
                    drop_temp_table,
                ) in finalized_query.temp_tables.values():
                    exit_stack.enter_context(
                        self._create_temp_table(
                            create_temp_table,
                            temp_table,
                            drop_temp_table,
                            sqlalchemy_cnx,
                            temp_value,
                        )
                    )
                with con.cursor() as cursor:
                    job_id = str(uuid.uuid4())
                    if query_execution_parameters.snowflake.snowflake_unload_stage is not None:
                        original_sql = sql
                        chalk_logger.info(
                            f"Executing query to unload data to Snowflake stage for unload {job_id=} {query_execution_parameters.snowflake.snowflake_unload_stage}"
                        )
                        sql = _rewrite_query_for_unload(
                            sql=sql,
                            unload_job_identifier=job_id,
                            snowflake_unload_stage=query_execution_parameters.snowflake.snowflake_unload_stage,
                        )

                    chalk_logger.info(f"Compiled query: {repr(sql)}")

                    res = cursor.execute(sql, named_params)
                    chalk_logger.info("Executed Snowflake query. Fetching results.")
                    assert res is not None

                    empty_batch_with_schema = None
                    if query_execution_parameters.snowflake.snowflake_unload_stage is not None:
                        # get the uris of the files that were unloaded
                        prefix = _stage_prefix(
                            job_id,
                            query_execution_parameters.snowflake.snowflake_unload_stage,
                        )
                        stage_list_cursor = cursor.execute(f"LIST {prefix}")
                        if stage_list_cursor is None:
                            chalk_logger.error("Failed to enumerate unloaded files in Snowflake stage")
                            raise ValueError("Failed to enumerate unloaded files in Snowflake stage")
                        stage_list = stage_list_cursor.fetchall()
                        chalk_logger.info(f"Found {len(stage_list)} files in Snowflake stage at prefix: {prefix}")
                        for row in stage_list:
                            # name/size/md5/last_modified
                            result_handles.put_nowait(
                                UnloadedStorageFileResultHandle(uri=row[0], compressed_size=row[1])
                            )
                        if cursor.rowcount == 0:
                            cursor.execute(original_sql, named_params)
                            chalk_logger.info(
                                "Executed Snowflake query to get empty batch with schema. Fetching results."
                            )
                            empty_batch_with_schema = (
                                cursor.fetch_arrow_all(True) if _has_new_fetch_arrow_all() else None
                            )
                    else:
                        chalk_logger.info("Fetching arrow tables from Snowflake.")
                        for batch in cursor.get_result_batches() or []:
                            result_handles.put_nowait(ResultBatchResultHandle(batch))

                        empty_batch_with_schema = (
                            cursor.fetch_arrow_all(True)
                            if cursor.rowcount == 0 and _has_new_fetch_arrow_all()
                            else None
                        )

        yield from self._yield_from_result_handles(
            result_handles, query_execution_parameters, columns_to_features, empty_batch_with_schema
        )

    def _yield_from_result_handles(
        self,
        result_handles: queue.Queue[ResultHandle],
        query_execution_parameters: QueryExecutionParameters,
        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
        empty_batch_with_schema: pa.Table | None,
    ):
        yielded = False
        max_weight = query_execution_parameters.max_prefetch_size_bytes
        if max_weight <= 0:
            max_weight = None
        pa_table_queue: queue.Queue[tuple[pa.Table, int] | _WorkerId] = queue.Queue()
        sem = None if max_weight is None else MultiSemaphore(max_weight)
        assert query_execution_parameters.num_client_prefetch_threads >= 1

        futures = {
            _WorkerId(i): self.executor.submit(
                self._download_worker,
                result_handles,
                sem,
                pa_table_queue,
                _WorkerId(i),
            )
            for i in range(query_execution_parameters.num_client_prefetch_threads)
        }
        schema: pa.Schema | None = None
        while len(futures) > 0:
            x = pa_table_queue.get()
            if isinstance(x, int):
                # It's a _WorkerId, meaning that this download worker is done
                # We'll pop this worker from the futures list, and then await the result
                # This will raise if the download worker crashed, which is what we want, or be a no-op if the download worker succeeded
                futures.pop(x).result()
                continue
            tbl, weight = x
            if schema is None:
                schema = tbl.schema
            try:
                if len(tbl) == 0:
                    continue
                assert isinstance(tbl, pa.Table)
                features = columns_to_features(tbl.schema.names)
                yield self._postprocess_table(features, tbl)
                safe_incr("chalk.snowflake.downloaded_bytes", tbl.nbytes or 0)
                safe_incr("chalk.snowflake.downloaded_rows", tbl.num_rows or 0)
                yielded = True
            finally:
                # Releasing the semaphore post-yield to better respect the limit
                if sem is not None and weight > 0:
                    sem.release(weight)
                    safe_set_gauge("chalk.snowflake.remaining_prefetch_bytes", sem.get_value())
        if not yielded and query_execution_parameters.yield_empty_batches:
            if schema is not None:
                features = columns_to_features(schema.names)
                yield pa.RecordBatch.from_arrays(
                    arrays=[[] for _ in features],
                    names=[x.root_fqn for x in features.values()],
                )
                return
            elif empty_batch_with_schema is not None:
                features = columns_to_features(empty_batch_with_schema.schema.names)
                yield self._postprocess_table(features, empty_batch_with_schema)

    @classmethod
    def register_sqlalchemy_compiler_overrides(cls):
        try:
            from chalk.sql._internal.integrations.snowflake_compiler_overrides import register_snowflake_compiler_hooks
        except ImportError:
            raise missing_dependency_exception("chalkpy[snowflake]")
        register_snowflake_compiler_hooks()

    def _recreate_integration_variables(self) -> dict[str, str]:
        return {
            k: v
            for k, v in [
                create_integration_variable(_SNOWFLAKE_ACCOUNT_ID_NAME, self.name, self.account_identifier),
                create_integration_variable(_SNOWFLAKE_WAREHOUSE_NAME, self.name, self.warehouse),
                create_integration_variable(_SNOWFLAKE_USER_NAME, self.name, self.user),
                create_integration_variable(_SNOWFLAKE_PASSWORD_NAME, self.name, self.password),
                create_integration_variable(_SNOWFLAKE_DATABASE_NAME, self.name, self.db),
                create_integration_variable(_SNOWFLAKE_SCHEMA_NAME, self.name, self.schema),
                create_integration_variable(_SNOWFLAKE_ROLE_NAME, self.name, self.role),
                create_integration_variable(_SNOWFLAKE_PRIVATE_KEY_B64_NAME, self.name, self.private_key_b64),
            ]
            if v is not None
        }
